package org.apache.bcel.classfile;

import java.util.Objects;
import java.util.Stack;
import java.util.stream.Stream;

public class DescendingVisitor implements Visitor {
   private final JavaClass clazz;
   private final Visitor visitor;
   private final Stack<Object> stack = new Stack();

   public DescendingVisitor(JavaClass clazz, Visitor visitor) {
      this.clazz = clazz;
      this.visitor = visitor;
   }

   private <E extends Node> void accept(E[] node) {
      Stream.of(node).forEach((e) -> {
         e.accept(this);
      });
   }

   public Object current() {
      return this.stack.peek();
   }

   public Object predecessor() {
      return this.predecessor(0);
   }

   public Object predecessor(int level) {
      int size = this.stack.size();
      return size >= 2 && level >= 0 ? this.stack.elementAt(size - (level + 2)) : null;
   }

   public void visit() {
      this.clazz.accept(this);
   }

   public void visitAnnotation(Annotations annotation) {
      this.stack.push(annotation);
      annotation.accept(this.visitor);
      this.accept(annotation.getAnnotationEntries());
      this.stack.pop();
   }

   public void visitAnnotationDefault(AnnotationDefault obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitAnnotationEntry(AnnotationEntry annotationEntry) {
      this.stack.push(annotationEntry);
      annotationEntry.accept(this.visitor);
      this.stack.pop();
   }

   public void visitBootstrapMethods(BootstrapMethods bm) {
      this.stack.push(bm);
      bm.accept(this.visitor);
      this.stack.pop();
   }

   public void visitCode(Code code) {
      this.stack.push(code);
      code.accept(this.visitor);
      this.accept(code.getExceptionTable());
      this.accept(code.getAttributes());
      this.stack.pop();
   }

   public void visitCodeException(CodeException ce) {
      this.stack.push(ce);
      ce.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantClass(ConstantClass constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantDouble(ConstantDouble constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantDynamic(ConstantDynamic obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantFieldref(ConstantFieldref constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantFloat(ConstantFloat constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantInteger(ConstantInteger constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantInterfaceMethodref(ConstantInterfaceMethodref constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantInvokeDynamic(ConstantInvokeDynamic constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantLong(ConstantLong constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantMethodHandle(ConstantMethodHandle obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantMethodref(ConstantMethodref constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantMethodType(ConstantMethodType obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantModule(ConstantModule obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantNameAndType(ConstantNameAndType constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantPackage(ConstantPackage obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantPool(ConstantPool cp) {
      this.stack.push(cp);
      cp.accept(this.visitor);
      Stream.of(cp.getConstantPool()).filter(Objects::nonNull).forEach((e) -> {
         e.accept(this);
      });
      this.stack.pop();
   }

   public void visitConstantString(ConstantString constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantUtf8(ConstantUtf8 constant) {
      this.stack.push(constant);
      constant.accept(this.visitor);
      this.stack.pop();
   }

   public void visitConstantValue(ConstantValue cv) {
      this.stack.push(cv);
      cv.accept(this.visitor);
      this.stack.pop();
   }

   public void visitDeprecated(Deprecated attribute) {
      this.stack.push(attribute);
      attribute.accept(this.visitor);
      this.stack.pop();
   }

   public void visitEnclosingMethod(EnclosingMethod obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitExceptionTable(ExceptionTable table) {
      this.stack.push(table);
      table.accept(this.visitor);
      this.stack.pop();
   }

   public void visitField(Field field) {
      this.stack.push(field);
      field.accept(this.visitor);
      this.accept(field.getAttributes());
      this.stack.pop();
   }

   public void visitInnerClass(InnerClass inner) {
      this.stack.push(inner);
      inner.accept(this.visitor);
      this.stack.pop();
   }

   public void visitInnerClasses(InnerClasses ic) {
      this.stack.push(ic);
      ic.accept(this.visitor);
      this.accept(ic.getInnerClasses());
      this.stack.pop();
   }

   public void visitJavaClass(JavaClass clazz) {
      this.stack.push(clazz);
      clazz.accept(this.visitor);
      this.accept(clazz.getFields());
      this.accept(clazz.getMethods());
      this.accept(clazz.getAttributes());
      clazz.getConstantPool().accept(this);
      this.stack.pop();
   }

   public void visitLineNumber(LineNumber number) {
      this.stack.push(number);
      number.accept(this.visitor);
      this.stack.pop();
   }

   public void visitLineNumberTable(LineNumberTable table) {
      this.stack.push(table);
      table.accept(this.visitor);
      this.accept(table.getLineNumberTable());
      this.stack.pop();
   }

   public void visitLocalVariable(LocalVariable var) {
      this.stack.push(var);
      var.accept(this.visitor);
      this.stack.pop();
   }

   public void visitLocalVariableTable(LocalVariableTable table) {
      this.stack.push(table);
      table.accept(this.visitor);
      this.accept(table.getLocalVariableTable());
      this.stack.pop();
   }

   public void visitLocalVariableTypeTable(LocalVariableTypeTable obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitMethod(Method method) {
      this.stack.push(method);
      method.accept(this.visitor);
      this.accept(method.getAttributes());
      this.stack.pop();
   }

   public void visitMethodParameter(MethodParameter obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitMethodParameters(MethodParameters obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      Stream.of(obj.getParameters()).forEach((e) -> {
         e.accept(this);
      });
      this.stack.pop();
   }

   public void visitModule(Module obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.accept(obj.getRequiresTable());
      this.accept(obj.getExportsTable());
      this.accept(obj.getOpensTable());
      this.accept(obj.getProvidesTable());
      this.stack.pop();
   }

   public void visitModuleExports(ModuleExports obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitModuleMainClass(ModuleMainClass obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitModuleOpens(ModuleOpens obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitModulePackages(ModulePackages obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitModuleProvides(ModuleProvides obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitModuleRequires(ModuleRequires obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitNestHost(NestHost obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitNestMembers(NestMembers obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitParameterAnnotation(ParameterAnnotations obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitParameterAnnotationEntry(ParameterAnnotationEntry obj) {
      this.stack.push(obj);
      obj.accept(this.visitor);
      this.stack.pop();
   }

   public void visitSignature(Signature attribute) {
      this.stack.push(attribute);
      attribute.accept(this.visitor);
      this.stack.pop();
   }

   public void visitSourceFile(SourceFile attribute) {
      this.stack.push(attribute);
      attribute.accept(this.visitor);
      this.stack.pop();
   }

   public void visitStackMap(StackMap table) {
      this.stack.push(table);
      table.accept(this.visitor);
      this.accept(table.getStackMap());
      this.stack.pop();
   }

   public void visitStackMapEntry(StackMapEntry var) {
      this.stack.push(var);
      var.accept(this.visitor);
      this.stack.pop();
   }

   public void visitSynthetic(Synthetic attribute) {
      this.stack.push(attribute);
      attribute.accept(this.visitor);
      this.stack.pop();
   }

   public void visitUnknown(Unknown attribute) {
      this.stack.push(attribute);
      attribute.accept(this.visitor);
      this.stack.pop();
   }
}
